@@ -49,6 +49,10 @@
 # include <netinet/ip_nat.h>
 #endif
 
+#ifdef IPTABLES
+# include <linux/netfilter_ipv4.h>
+#endif
+
 #ifdef TCP_WRAPPERS
 # include <tcpd.h>
 #endif
@@ -117,6 +121,7 @@ typedef struct connection_
 	buffer_t			read_buffer;
 	buffer_t			request;
 	int					parse_state;
+	int					method_end_offset;
 	int					url_offset;
 	int					url_end_offset;
 	int					version_offset;
@@ -167,6 +172,7 @@ static void log_it(char *fmt, ...);
 static char				*prog;
 static int				daemonize = 1;
 static int				fully_transparent = 0;
+static int				proxy_only = 0;
 static char				*force_url = NULL;
 static int				force_url_length;
 #ifdef LOG_TO_FILE
@@ -225,12 +231,18 @@ int main(int argc, char **argv)
 	/*
 	 * Parse the command line arguments.
 	 */
-	while ((arg = getopt(argc, argv, "dts:r:b:f:l:a:")) != EOF)
+	while ((arg = getopt(argc, argv, "dtps:r:b:f:l:a:")) != EOF)
 	{
 		switch (arg)
 		{
 		case 't':
 			fully_transparent = 1;
+			proxy_only = 0;
+			break;
+
+		case 'p':
+			proxy_only = 1;
+			fully_transparent = 0;
 			break;
 
 		case 's':
@@ -526,11 +538,12 @@ static void usage(char *prog, char *opt)
 		fprintf(stderr, "%s: %s\n", prog, opt);
 	}
 #ifdef LOG_TO_FILE
-	fprintf(stderr, "usage: %s [-t] [-f url] [-s port [-d] [-b ipaddr] [-r user] [-a acl]] [-l file] proxyhost proxyport\n", prog);
+	fprintf(stderr, "usage: %s [-t | -p] [-f url] [-s port [-d] [-b ipaddr] [-r user] [-a acl]] [-l file] proxyhost proxyport\n", prog);
 #else
-	fprintf(stderr, "usage: %s [-t] [-f url] [-s port [-d] [-b ipaddr] [-r user] [-a acl]] proxyhost proxyport\n", prog);
+	fprintf(stderr, "usage: %s [-t | -p] [-f url] [-s port [-d] [-b ipaddr] [-r user] [-a acl]] proxyhost proxyport\n", prog);
 #endif
 	fprintf(stderr, "    -t          Operate transparently, don't talk to a proxy.\n");
+	fprintf(stderr, "    -p          Only talk to a proxy, disable direct fall-back.\n");
 	fprintf(stderr, "    -f url      Force access to always go to specified URL.\n");
 	fprintf(stderr, "    -s port     Run as a server bound to the specified port.\n");
 	fprintf(stderr, "    -d          Do not background the daemon in server mode.\n");
@@ -701,6 +714,7 @@ static int bind_to_port(ip_address_t bind_ip, short bind_port)
 	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
 	{
 		perror("bind()");
+		close(sock);
 		exit(1);
 	}
 
@@ -710,6 +724,7 @@ static int bind_to_port(ip_address_t bind_ip, short bind_port)
 	if (listen(sock, SOMAXCONN) < 0)
 	{
 		perror("listen()");
+		close(sock);
 		exit(1);
 	}
 
@@ -751,6 +766,7 @@ static int connect_to_proxy(ip_address_t ip, short port)
 #if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 		syslog(LOG_WARNING, "connect(): %m");
 #endif
+		close(sock);
 		return (-1);
 	}
 
@@ -761,6 +777,7 @@ static int connect_to_proxy(ip_address_t ip, short port)
 		int		on = 1;
 		setsockopt(STDIN_FILENO, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof(on));
 	}
+
 	return (sock);
 }
 
@@ -848,7 +865,7 @@ static void server_main_loop(int sock, char *server_hostname, short server_port)
 		sigaction(SIGCHLD, &sa, NULL);
 	}
 #else
-	/* Go Linux */
+	/* Go Linux & SVR4 */
  	signal(SIGCHLD, SIG_IGN);
 #endif
 
@@ -996,6 +1013,24 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 	conn.parse_state = PS_METHOD;
 	conn.parse_state = PS_METHOD;
 
+#ifdef IPTABLES
+	/*
+	 * The first thing we do is get the IP address that the client was
+	 * trying to connected to. Here lies part of the magic. Linux-2.4
+	 * introduces IPTABLES which uses getsockopt instead of getsockname
+	 * for a more obvious interface.
+	 */
+	length = sizeof(conn.dest_addr);
+	if (getsockopt(sock, SOL_IP, SO_ORIGINAL_DST, (char *)&conn.dest_addr, &length) < 0)
+	{
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+		syslog(LOG_ERR, "getsockopt(SO_ORIGINAL_DST): %m");
+#endif
+		return;
+	}
+	
+#else/*!IPTABLES*/
+
 	/*
 	 * The first thing we do is get the IP address that the client was
 	 * trying to connected to. Here lies part of the magic. Normally
@@ -1012,10 +1047,10 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 
 #ifdef IPFILTER
 	/*
-	 * However, under IPFILTER, we have to do more work. The "to_addr" has
-	 * been changed by the IP filter to trick the IP stack into processing
-	 * the packet locally. We need to lookup the real remote address
-	 * ourselves using an ioctl().
+	 * However, under IPFILTER, we have to do additional work. The "to_addr"
+	 * has been changed by the IP filter to trick the IP stack into
+	 * processing the packet locally. We need to lookup the real remote
+	 * address ourselves using an ioctl().
 	 */
 
 	/*
@@ -1043,6 +1078,8 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 	conn.dest_addr.sin_port = natlook.nl_realport;
 #endif
 
+#endif/*!IPTABLES*/
+
 	/*
 	 * Lookup proxy server IP address. Because we do the lookup for every
 	 * request we make use of DNS round-robin for load balancing.
@@ -1063,16 +1100,25 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 	if (fully_transparent ||
 		(conn.proxy_fd = connect_to_proxy(server_ip, server_port)) == -1)
 	{
-		conn.proxy_fd = connect_to_proxy(conn.dest_addr.sin_addr.s_addr, conn.dest_addr.sin_port);
-		fully_transparent = 1;
+		if (!proxy_only)
+		{
+			conn.proxy_fd = connect_to_proxy(conn.dest_addr.sin_addr.s_addr, conn.dest_addr.sin_port);
+			fully_transparent = 1;
+		}
+		else
+		{
+			conn.proxy_fd = -1;
+		}
 	}
 	if (conn.proxy_fd == -1)
 	{
 #if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 		if (fully_transparent)
 			syslog(LOG_ERR, "Could not connect to destination");
+		else if (proxy_only)
+			syslog(LOG_ERR, "Could not connect to proxy host \"%s\"", server_hostname);
 		else
-			syslog(LOG_ERR, "Could not connect to cache host: %s", server_hostname);
+			syslog(LOG_ERR, "Could not connect to proxy host \"%s\" or destination", server_hostname);
 #endif
 		return;
 	}
@@ -1275,7 +1321,9 @@ static int process_client_request(connection_t *conn, size_t last_read)
 		case PS_METHOD:
 			if (c == '\r')					conn->parse_state = PS_METHOD + 1;
 			else if (c == '\n')				conn->parse_state = PS_END_OF_REQUEST;
-			else if (c == ' ')				conn->parse_state = PS_METHOD_END;
+			else if (c == ' ') {			conn->parse_state = PS_METHOD_END;
+				conn->method_end_offset = index;
+			}
 			break;
 
 		case PS_METHOD + 1:
@@ -1656,13 +1704,15 @@ static int process_client_request(connection_t *conn, size_t last_read)
 			 * Build the resulting request into "request_buffer" and set "send_size"
 			 * to the length of the request. This enables the final request to be
 			 * written with a single write(2) system call. This should be more
-			 * efficient for both transproxy and the cache.
+			 * efficient for both transproxy and the proxy.
 			 */
-			if (fully_transparent)
+			if ((fully_transparent) ||
+				(strncasecmp(conn->request.buffer, "CONNECT", 7) == 0))
 			{
 				/*
 				 * Leave the request as it is. Transparent mode is usefull for
-				 * operation in a proxy firewall.
+				 * operation in a proxy firewall. Or HTTPS protocol doesn't
+				 * like modifications so just pass the request on.
 				 */
 				memcpy(request_buffer, conn->request.buffer, index);
 				send_size = index;
